local get_id = QBL.util.Links.get_id

-- Enabled by LOOT_OPENED in lib/Loot and disabled by LOOT_CLOSED in this file.
if QBL.var.is_looting == nil then QBL.var.is_looting = false end

-- Toggled by stackup, stack_queue and BANKFRAME_OPENED.
-- Prevents concurrent runs of the stacking routines.
local is_stacking = false

-- Frame used to register timer events for stacking.
local w_frame = getglobal("WorldFrame")

-- Ran before general stacking and after stacking has complete, this routine
-- will scan for special bags, then place compatible items inside each bag from
-- the player inventory.
local function stack_special_bags()
	local special, bags = {}, {}

	-- Scan for special bags with free space.
	for bag = BACKPACK_CONTAINER + 1, NUM_BAG_SLOTS do
		-- If not an ignored bag...
		if not GetBagSlotFlag(bag, 1) then
			local free, type = GetContainerNumFreeSlots(bag)
			if free > 0 and type > 0 then
				table.insert(special, {
					id = ContainerIDToInventoryID(bag),
					idx = bag, type = type
				})
			-- Collect normal bags to stack from.
			elseif type == 0 then
				table.insert(bags, {idx = bag, type = 0})
			end
		end
	end

	-- Include the backpack, if not ignored.
	if not GetBackpackAutosortDisabled()
	then table.insert(bags, {idx = BACKPACK_CONTAINER, type = 0}) end

	-- For every special bag...
	for bag = 1, #special do
		-- Compare against all elligible bags...
		for next = 1, #bags do
			if special[bag].type ~= bags[next].type then
				for slot = 1, GetContainerNumSlots(bags[next].idx) do
					local item = get_id(GetContainerItemLink(bags[next].idx, slot))
					if item and select(9, GetItemInfo(item)) ~= "INVTYPE_BAG" then
						local type = GetItemFamily(item)
						if type and type > 0 then
							local free = GetContainerNumFreeSlots(special[bag].idx)
							-- Item types are bit masks.
							if bit.band(type, special[bag].type) == special[bag].type
							then
								PickupContainerItem(bags[next].idx, slot)
								PutItemInBag(special[bag].id)
								free = free - 1
							end
							if free <= 0 then break end
						end
					end
				end
			end
		end
	end
end

-- Processes over 'queue' task with each OnUpdate event, taking a minimum of
-- three calls to perform the stack 'task'. Each event branches on a different
-- 'state'. Upon performing the branch, the 'state' of the calling object is
-- incremented. The third state is to increment queue index 'idx'. When the
-- queue is finished, stack_queue is removed from OnUpdate events.
local elapsed, stack_idx, stack_state = 0, 0, 0
local function stack_queue(self, queue, type)
	local task = queue[stack_idx]
	if stack_state == 1 then
		if type == "gbank" then
			if task.split
			then SplitGuildBankItem(task.oldbag, task.oldslot, task.count)
			else PickupGuildBankItem(task.oldbag, task.oldslot) end
		else
			if task.split
			then SplitContainerItem(task.oldbag, task.oldslot, task.count)
			else PickupContainerItem(task.oldbag, task.oldslot) end
		end
		stack_state = 2
	elseif stack_state == 2 then
		elapsed = elapsed + 1
		if elapsed > 3 then
			elapsed = 0
			PickupContainerItem(task.newbag, task.newslot)
			stack_state = 3
		end
	elseif stack_state == 3 then
		stack_state = 1
		stack_idx = stack_idx + 1
		if queue[stack_idx] == nil then
			self:SetScript("OnUpdate", nil)
			stack_special_bags()
			stack_idx = 0
			stack_state = 0
			is_stacking = false
		end
	end
end

-- Translates an 'items' table of 'type' into a 'queue' of movement
-- operations to perform. The queue is then cleared with stack_queue via
-- OnUpdate events.
local function make_queue(items, type)
	local queue = {}

	-- Compare each item with every other item in the table.
	-- Items are removed from consideration by setting them to { id = -1 }.
	for first = 1, #items do
		for next = 1, #items do
			if next ~= first
				and not (items[first].id == -1 or items[next].id == -1)
				and items[next].id == items[first].id
				and items[first].need <= items[next].need
			then
				if items[next].count < items[first].need then
					table.insert(queue, {split = false, oldbag = items[next].bag,
						oldslot = items[next].slot, newbag = items[first].bag,
						newslot = items[first].slot})
					items[first].count = items[first].count + items[next].count
					items[first].need = items[first].stack - items[first].count
					items[next] = {id = -1}
				elseif items[next].count == items[first].need then
					table.insert(queue, {split = false, oldbag = items[next].bag,
						oldslot = items[next].slot, newbag = items[first].bag,
						newslot = items[first].slot})
					items[first] = {id = -1}
					items[next] = {id = -1}
				elseif items[next].count > items[first].need then
					table.insert(queue, {split = true, count = items[first].need,
						oldbag = items[next].bag, oldslot = items[next].slot,
						newbag = items[first].bag, newslot = items[first].slot})
					items[next].count = items[next].count - items[first].need
					items[next].need = items[next].stack - items[next].count
					items[first] = {id = -1}
				end
			end
		end
	end

	if not (next(queue) == nil) then
		stack_idx, stack_state = 1, 1
		w_frame:SetScript("OnUpdate", function(self)
			stack_queue(self, queue, type) end)
	else is_stacking = false end
end

-- Scan 'bag' of 'type' and return a table of stackable items inside it.
local function make_tbl(bag, type)
	local tbl = {}
	if type == "gbank" then
		for slot = 1, 98 do
			local _, count = GetGuildBankItemInfo(bag, slot)
			if count > 0 then
				local id = get_id(GetGuildBankItemLink(bag, slot))
				if id then
					local item, _, _, _, _, _, _, stack = GetItemInfo(id)
					if count < stack then
						table.insert(tbl, {id = id, bag = bag, slot = slot,
							count = count, stack = stack, need = stack - count})
					end
				end
			end
		end
	else
		for slot = 1, GetContainerNumSlots(bag) do
			local _, count, _, quality = GetContainerItemInfo(bag, slot)
			if quality == -1 or count then -- -1 are unique items that may stack.
				local id = get_id(GetContainerItemLink(bag, slot))
				if id then
					local stack = select(8, GetItemInfo(id))
					if stack and count < stack then
						table.insert(tbl, {id = id, bag = bag, slot = slot,
							count = count, stack = stack, need = stack - count})
					end
				end
			end
		end
	end
	if not (next(tbl) == nil) then return tbl end
end

-- Returns a boolean table where the index is the position of an elligible bag
-- for stacking.
local function count_bags(type)
	local bags = {}
	if type == "gbank" then
		tabs = GetNumGuildBankTabs()
		for tab = 1, tabs do
			local deposit = select(4, GetGuildBankTabInfo(tab))
			if IsGuildLeader(UnitName("player")) then deposit = true end
			if deposit then bags[tab] = true end
		end
	else
		local first, last
		if type == "bank" then
			first = ITEM_INVENTORY_BANK_BAG_OFFSET + 1
			last = ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS
			if not GetBagSlotFlag(BANK_CONTAINER, 1)
			then bags[BANK_CONTAINER] = true end
		else -- is "bags"
			first, last = BACKPACK_CONTAINER + 1, NUM_BAG_SLOTS
			-- Include the backpack, if not ignored.
			if not GetBackpackAutosortDisabled()
			then first = BACKPACK_CONTAINER end
		end
		for bag = first, last do
			-- If not an ignored bag and it has slots...
			if not GetBagSlotFlag(bag, 1)
				and GetContainerNumSlots(bag) > 0
			then bags[bag] = true end
		end
	end
	if not (next(bags) == nil) then return bags end
end

-- Called by stacking events with the 'type' of content to stack.
-- 'type' may be 'bags', 'bank' or 'gbank'.
local function stackup(type)
	if QBL_cfg.stacker
		and not is_stacking
		and not CursorHasItem()
		and not CursorHasMoney()
		and not CursorHasSpell()
	then
		if type == "gbank" and not QBL_cfg.stack_gbank then return end
		is_stacking = true

		if type == "bags" then stack_special_bags() end

		local bags = count_bags(type)
		if not bags then
			is_stacking = false
			return
		end

		local items = {}
		for bag in pairs(bags) do
			local t = make_tbl(bag, type)
			if t then
				for _, v in ipairs(t) do table.insert(items, v) end
			end
		end
		if not (next(items) == nil)
		then make_queue(items, type) else is_stacking = false end
	end
end

function QBL.events:BAG_UPDATE(...)
	if not QBL.var.is_looting
		and not TradeFrame:IsVisible()
		and not SendMailFrame:IsVisible()
	then
		if AuctionFrame and AuctionFrame:IsVisible() then return
		elseif BankFrame:IsVisible() then stackup("bank")
		elseif GuildBankFrame and GuildBankFrame:IsVisible()
		then stackup("gbank") end
		stackup("bags")
	end
end

-- These global hook hacks are needed to safely call DepositReagentBank
-- without first opening the reagent bank window.
local deposit_hook = false
function QBL.events:BANKFRAME_OPENED(...)
	if QBL_cfg.reagent then
		is_stacking = true
		if not _BankFrameItemButton_Update then
			_BankFrameItemButton_Update = BankFrameItemButton_Update
			BankFrameItemButton_Update = function(button)
				if deposit_hook == false and button ~= nil
				then _BankFrameItemButton_Update(button)
				elseif button == nil then -- Do nothing.
				else deposit_hook = false end
			end
		end
		deposit_hook = true
		DepositReagentBank()
		is_stacking = false
	end
	-- We want to support turning off 'stackup' but running 'reagent'.
	if QBL_cfg.stacker then stackup("bank") end
end

function QBL.events:LOOT_CLOSED(...)
	QBL.var.is_looting = false
	stackup("bags")
end

function QBL.events:AUCTION_HOUSE_CLOSED(...) stackup("bags") end
function QBL.events:GUILDBANKBAGSLOTS_CHANGED(...) stackup("gbank") end
function QBL.events:GUILDBANKFRAME_OPENED(...) stackup("gbank") end
function QBL.events:MAIL_CLOSED(...) stackup("bags") end
function QBL.events:PLAYERBANKSLOTS_CHANGED(...) stackup("bank") end
function QBL.events:TRADE_CLOSED(...) stackup("bags") end

QBL.srv.stacker = {"AUCTION_HOUSE_CLOSED", "BAG_UPDATE", "BANKFRAME_OPENED",
	"GUILDBANKBAGSLOTS_CHANGED", "GUILDBANKFRAME_OPENED", "LOOT_CLOSED",
	"MAIL_CLOSED", "PLAYERBANKSLOTS_CHANGED", "TRADE_CLOSED"}
QBL.srv.reagent = {"BANKFRAME_OPENED"}
